"use strict";

function _typeof(obj) { "@babel/helpers - typeof"; if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

function _createForOfIteratorHelper(o, allowArrayLike) { var it; if (typeof Symbol === "undefined" || o[Symbol.iterator] == null) { if (Array.isArray(o) || (it = _unsupportedIterableToArray(o)) || allowArrayLike && o && typeof o.length === "number") { if (it) o = it; var i = 0; var F = function F() {}; return { s: F, n: function n() { if (i >= o.length) return { done: true }; return { done: false, value: o[i++] }; }, e: function e(_e) { throw _e; }, f: F }; } throw new TypeError("Invalid attempt to iterate non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); } var normalCompletion = true, didErr = false, err; return { s: function s() { it = o[Symbol.iterator](); }, n: function n() { var step = it.next(); normalCompletion = step.done; return step; }, e: function e(_e2) { didErr = true; err = _e2; }, f: function f() { try { if (!normalCompletion && it["return"] != null) it["return"](); } finally { if (didErr) throw err; } } }; }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

(function () {
  //@ts-check
  'use strict';
  /**
   * @typedef {string|RegExp|RegExpTemplate} TemplateElement
   */

  /**
   * @typedef {TemplateElement|TemplateNode} TemplateInternal
   */

  /**
   * @typedef {Object<string, Set<number>>} VarIndexMap
   */

  /**
   * @typedef {Map<TemplateNode, Set<number>>} ChildrenIndexMap
   */

  /**
   * !!! internal helper function !!!
   * Returns the content string.
   * i.e. the RegExp without slashes notation and flags.
   * @param {RegExp} regexp
   * @returns {string} the processed regexp.
   */

  function getRegExpPattern(regexp) {
    // convert to string
    var reStr = regexp.toString(); // find the index of the last slash

    var endIndex = reStr.length;

    while (reStr[endIndex] !== '/') {
      endIndex--;
    } // return the regexp pattern


    return reStr.substring(1, endIndex);
  }
  /**
   * !!! internal helper function !!!
   *
   * @param {string} regExpPattern pattern to
   * search and slit by vars
   *
   * @returns {string[]} an array with
   * all parts separated by the found variables.
   * The odd positions contain the variables' names.
   */


  function splitVars(regExpPattern) {
    return regExpPattern.split(/\\V\s*\{\s*([a-zA-Z]\w*)\s*\}/g);
  }
  /**
   * !!! internal function !!!
   * Escape all especial chars
   * i.e. all chars should be literals.
   * note: special string chars such
   * as \n are already interpreted before
   * this function is called, and for
   * that reason, not escaped.
   *
   * @param {string} str input string
   * @returns {string} escaped string
   */


  function escapeString(str) {
    return str // escape back slash
    .replace(/\\/g, '\\\\') // escape special chars
    .replace(/[+*?.^$(){}[\]|]/g, function (_char) {
      return '\\' + _char;
    });
  }
  /**
   * !!! internal helper function !!!
   * Process regexp for template.
   * Get pattern, create node if necessary,
   * add node and process variables.
   *
   * @param {RegExp} regexp RegExp to be processed.
   *
   * @param {TemplateNode[]} templateStack
   * the target RegExpTemplate's templateStack.
   *
   * @returns {string|TemplateNode} the processed regexp.
   */


  function processRegExp(regexp, templateStack) {
    var regExpPattern = getRegExpPattern(regexp);
    var varSplit = splitVars(regExpPattern); // if a variable is found then
    // the varSplit array's length is >= 3

    if (varSplit.length >= 3) {
      var node = new TemplateNode(templateStack.length);
      node._body = varSplit;
      templateStack.push(node);
      return node;
    }

    return varSplit[0];
  }
  /**
   * !!! internal helper function !!!
   * Process input element to an valid TemplateInternal.
   *
   * @param {*} element element to process.
   *
   *
   * @param {TemplateNode[]} templateStack
   * the target RegExpTemplate's templateStack.
   *
   * @returns {TemplateInternal}
   */


  function processElement(element, templateStack) {
    if (element instanceof RegExp) {
      return processRegExp(element, templateStack);
    } else if (element instanceof RegExpTemplate) {
      return element;
    } else {
      return escapeString("" + element);
    }
  }
  /**
   * !!! internal class !!!
   * Used to represent the template internal tree.
   */


  var TemplateNode = /*#__PURE__*/function () {
    /**
     * @param {number} index index of the node in the templateStack.
     */
    function TemplateNode(index) {
      _classCallCheck(this, TemplateNode);

      this._index = index;
      /**
       * Store all node's template elements
       * and/or children nodes.
       * @type {TemplateInternal[]}
       */

      this._body = [];
      /**
       * Set with all the parent nodes
       * @type {Set<TemplateNode>}
       */

      this._parents = null;
      /**
       * Maps of children to they positions
       * in this node's body.
       * @type {ChildrenIndexMap}
       */

      this._children = new Map();
      /**
       * Stores all variables' positions
       * in the node.
       * @type {VarIndexMap}
       */

      this._vars = null;
    }
    /**
     * Adds elements or nodes to the node's body.
     * Also keeps track of the children nodes positions
     * @param {...TemplateInternal} elements
     */


    _createClass(TemplateNode, [{
      key: "add",
      value: function add() {
        var _iterator = _createForOfIteratorHelper(arguments),
            _step;

        try {
          for (_iterator.s(); !(_step = _iterator.n()).done;) {
            var element = _step.value;

            // if the element is a TemplateNode
            // then map its position in this._children
            if (element instanceof TemplateNode) {
              // init Set if necessary for element in
              // this node's children map.
              var childMap = this._children.get(element);

              if (!childMap) {
                childMap = new Set();

                this._children.set(element, childMap);
              }

              childMap.add(this._body.length);
            }

            this._body.push(element);
          }
        } catch (err) {
          _iterator.e(err);
        } finally {
          _iterator.f();
        }
      }
      /**
       * Maps all variables in the node
       * into they positions in the node's body
       */

    }, {
      key: "mapVariables",
      value: function mapVariables() {
        /**
         * @type {VarIndexMap}
         */
        var vars = {};

        for (var i = 1; i < this._body.length; i += 2) {
          var escapedVar = ">" + this._body[i];

          if (escapedVar in vars) {
            vars[escapedVar].add(i);
          } else {
            var varSet = new Set();
            varSet.add(i);
            vars[escapedVar] = varSet;
          }
        }

        this._vars = vars;
      }
      /**
       * Propagates the vars of the node
       * to all the nodes above.
       */

    }, {
      key: "propagateVars",
      value: function propagateVars() {
        var _this = this;

        // stop if is the root node
        if (!this._parents) return;

        var _iterator2 = _createForOfIteratorHelper(this._parents),
            _step2;

        try {
          var _loop = function _loop() {
            var parent = _step2.value;
            var selfVarNames = Object.getOwnPropertyNames(_this._vars);

            if (selfVarNames.length > 0 && !parent._vars) {
              parent._vars = {};
            }
            /**
             * After adding the new positions, this boolean
             * will decide if current parent should propagate as well.
             * It will happen if at least one variable is not defined.
             * @type {boolean}
             */


            var shouldPropagate = false;

            var _iterator3 = _createForOfIteratorHelper(selfVarNames),
                _step3;

            try {
              var _loop2 = function _loop2() {
                var varName = _step3.value;

                if (!(varName in parent._vars)) {
                  parent._vars[varName] = new Set();
                  shouldPropagate = true;
                } // the parent var positions are the ones there before
                // plus the positions of the child node propagating


                parent._children.get(_this).forEach(function (index) {
                  parent._vars[varName].add(index);
                });
              };

              for (_iterator3.s(); !(_step3 = _iterator3.n()).done;) {
                _loop2();
              }
            } catch (err) {
              _iterator3.e(err);
            } finally {
              _iterator3.f();
            }

            if (shouldPropagate) {
              parent.propagateVars();
            }
          };

          for (_iterator2.s(); !(_step2 = _iterator2.n()).done;) {
            _loop();
          }
        } catch (err) {
          _iterator2.e(err);
        } finally {
          _iterator2.f();
        }
      }
      /**
       * Replaces the variable with the varValue given
       * for the current node and all its subnodes.
       * When replaced, the variable will be removed
       * from the current node and in its subnodes.
       *
       * @param {*} varName name of the variable to replace
       * @param {*} varValue value to replace in the variable
       * @param {Set<TemplateNode>} alreadyVisitedNodes set to keep
       * track of the nodes that the replaceVar was already called
       * @param {Map<TemplateNode, Set<number>>} varIndexes
       * indexes of the leaves of the variable
       * @see RegExpTemplate.prototype.applyVars
       */

    }, {
      key: "replaceVar",
      value: function replaceVar(varName, varValue, alreadyVisitedNodes, varIndexes) {
        var _this2 = this;

        // if the variable was already replaced
        // for this node then ignore it
        if (alreadyVisitedNodes.has(this)) {
          return;
        } else {
          alreadyVisitedNodes.add(this);
        }
        /**
         * Set to add the indexes of leaves (if exist)
         * of the variable.
         * @type {Set<number>}
         */


        var currentNodeVarIndexes = new Set();

        this._vars[">" + varName].forEach(function (index) {
          var element = _this2._body[index];

          if (element instanceof TemplateNode) {
            element.replaceVar(varName, varValue, alreadyVisitedNodes, varIndexes);
          } else {
            currentNodeVarIndexes.add(index);
            _this2._body[index] = varValue;
          }
        }); // only add the node to varIndexes if the
        // variable index points at least to one leave.
        // this will garanty that all the keys are parents
        // of the generated node (if generated) to replace.


        if (currentNodeVarIndexes.size > 0) {
          varIndexes.set(this, currentNodeVarIndexes);
        }

        delete this._vars[">" + varName];
      }
      /**
       * Compiles all the content of the node to a string.
       * @param {*} compiledNodes
       * @returns {string} the compiled string.
       */

    }, {
      key: "compile",
      value: function compile(compiledNodes) {
        var result = "";

        for (var i = 0; i < this._body.length; i++) {
          var currentElement = this._body[i];

          if (currentElement instanceof RegExpTemplate) {
            currentElement = getRegExpPattern(currentElement.compile());
          } else if (currentElement instanceof TemplateNode) {
            currentElement = compiledNodes[currentElement._index];
          }

          result += currentElement;
        }

        return result;
      }
    }]);

    return TemplateNode;
  }();

  var RegExpTemplate = /*#__PURE__*/function () {
    /**
     * @param {...*} elements - starting elements to be added to the template.
     */
    function RegExpTemplate() {
      _classCallCheck(this, RegExpTemplate);

      var rootNode = new TemplateNode(0);
      /**
       * @private
       * @type {TemplateNode[]}
       */

      this._templateStack = [rootNode];

      for (var _len = arguments.length, elements = new Array(_len), _key = 0; _key < _len; _key++) {
        elements[_key] = arguments[_key];
      }

      for (var _i = 0, _elements = elements; _i < _elements.length; _i++) {
        var rawElement = _elements[_i];
        var element = processElement(rawElement, this._templateStack);
        rootNode.add(element); // if the new element is a new node

        if (element instanceof TemplateNode) {
          element._parents = new Set();

          element._parents.add(rootNode);

          element.mapVariables();
          element.propagateVars();
        }
      }
    }
    /**
     * Replace the key variables in template,
     * with the values passed.
     * @param {*} vars variables to replace
     * @returns {RegExpTemplate} returns the template it self for chaining.
     */


    _createClass(RegExpTemplate, [{
      key: "applyVars",
      value: function applyVars(vars) {
        var _this3 = this;

        {
          var _iterator4 = _createForOfIteratorHelper(Object.getOwnPropertyNames(vars)),
              _step4;

          try {
            var _loop3 = function _loop3() {
              var varName = _step4.value;
              // process var value
              var varValue = processElement(vars[varName], _this3._templateStack);
              /**
               * If the varValue creates a new node, this
               * map will register the positions of the old var
               * positions to setup later the "parents" in the new node
               * and to add the new node to all its parents' "children"
               * @type {Map<TemplateNode, Set<number>>}
               */

              var varIndexes = new Map();
              /**
               * keeps track of the nodes that
               * already have been called replaceVar
               * @type {Set<TemplateNode>}
               */

              var alreadyVisitedNodes = new Set(); // replace var with the processed value

              _this3._templateStack[0].replaceVar(varName, varValue, alreadyVisitedNodes, varIndexes);

              if (varValue instanceof TemplateNode) {
                // init new node's "parents" and
                // map its positions in parents' "children"
                varValue._parents = new Set();
                varIndexes.forEach(function (indexSet, parent) {
                  varValue._parents.add(parent);

                  parent._children.set(varValue, indexSet);
                });
                varValue.mapVariables();
                varValue.propagateVars();
              }
            };

            for (_iterator4.s(); !(_step4 = _iterator4.n()).done;) {
              _loop3();
            }
          } catch (err) {
            _iterator4.e(err);
          } finally {
            _iterator4.f();
          }
        }
        return this;
      }
    }, {
      key: "compile",

      /**
       * Compiles all the template elements to a single RegExp.
       * @returns {RegExp} the compiled RegExp
       */
      value: function compile() {
        /**
         * @type {string[]}
         */
        var compiledNodes = new Array(this._templateStack.length);

        if (this._templateStack[0]._vars) {
          var unassignedVars = Object.getOwnPropertyNames(this._templateStack[0]._vars);

          if (unassignedVars.length > 0) {
            throw TypeError("Cannot compile due to unassigned variables.");
          }
        }

        for (var i = this._templateStack.length - 1; i >= 0; i--) {
          compiledNodes[i] = this._templateStack[i].compile(compiledNodes);
        }

        return new RegExp(compiledNodes[0]);
      }
    }, {
      key: "unassignedVars",
      get: function get() {
        return Object.getOwnPropertyNames(this._templateStack[0]._vars).map(function (v) {
          return v.substring(1);
        }); // remove escape char
      }
    }]);

    return RegExpTemplate;
  }();

  if ((typeof module === "undefined" ? "undefined" : _typeof(module)) === "object") {
    // export for node
    module.exports = RegExpTemplate;
  } else {
    // export for browser
    window["getRegExpPattern"] = RegExpTemplate;
  }
})();
//# sourceMappingURL=regexp-template.js.map
